@@ -1,7 +1,7 @@
 /*
  * WPA Supplicant - driver interaction with MADWIFI 802.11 driver
  * Copyright (c) 2004, Sam Leffler <sam@errno.com>
- * Copyright (c) 2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -26,6 +26,7 @@
 #include "eloop.h"
 #include "wpa_supplicant.h"
 #include "wpa.h"
+#include "wireless_copy.h"
 
 #include <include/compat.h>
 #include <net80211/ieee80211.h>
@@ -37,8 +38,10 @@
 #include <net80211/ieee80211_crypto.h>
 #include <net80211/ieee80211_ioctl.h>
 
-#include <net/if_arp.h>
-#include <linux/wireless.h>
+#ifdef IEEE80211_IOCTL_SETWMMPARAMS
+/* Assume this is built against madwifi-ng */
+#define MADWIFI_NG
+#endif /* IEEE80211_IOCTL_SETWMMPARAMS */
 
 struct wpa_driver_madwifi_data {
 	void *wext; /* private data for driver_wext */
@@ -72,6 +75,32 @@ set80211priv(struct wpa_driver_madwifi_data *drv, int op, void *data, int len,
 
 	if (ioctl(drv->sock, op, &iwr) < 0) {
 		if (show_err) {
+#ifdef MADWIFI_NG
+			int first = IEEE80211_IOCTL_SETPARAM;
+			int last = IEEE80211_IOCTL_GETMODE;
+			static const char *opnames[] = {
+				"ioctl[IEEE80211_IOCTL_SETPARAM]",
+				"ioctl[IEEE80211_IOCTL_GETPARAM]",
+				"ioctl[IEEE80211_IOCTL_SETKEY]",
+				"ioctl[IEEE80211_IOCTL_SETWMMPARAMS]",
+				"ioctl[IEEE80211_IOCTL_DELKEY]",
+				"ioctl[IEEE80211_IOCTL_GETWMMPARAMS]",
+				"ioctl[IEEE80211_IOCTL_SETMLME]",
+				"ioctl[IEEE80211_IOCTL_GETCHANINFO]",
+				"ioctl[IEEE80211_IOCTL_SETOPTIE]",
+				"ioctl[IEEE80211_IOCTL_GETOPTIE]",
+				"ioctl[IEEE80211_IOCTL_ADDMAC]",
+				NULL,
+				"ioctl[IEEE80211_IOCTL_DELMAC]",
+				"ioctl[IEEE80211_IOCTL_GETCHANLIST]",
+				"ioctl[IEEE80211_IOCTL_SETCHANLIST]",
+				NULL,
+				"ioctl[IEEE80211_IOCTL_SETMODE]",
+				"ioctl[IEEE80211_IOCTL_GETMODE]",
+			};
+#else /* MADWIFI_NG */
+			int first = IEEE80211_IOCTL_SETPARAM;
+			int last = IEEE80211_IOCTL_CHANLIST;
 			static const char *opnames[] = {
 				"ioctl[IEEE80211_IOCTL_SETPARAM]",
 				"ioctl[IEEE80211_IOCTL_GETPARAM]",
@@ -89,9 +118,9 @@ set80211priv(struct wpa_driver_madwifi_data *drv, int op, void *data, int len,
 				NULL,
 				"ioctl[IEEE80211_IOCTL_CHANLIST]",
 			};
-			if (IEEE80211_IOCTL_SETPARAM <= op &&
-			    op <= IEEE80211_IOCTL_CHANLIST)
-				perror(opnames[op - SIOCIWFIRSTPRIV]);
+#endif /* MADWIFI_NG */
+			if (first <= op && op <= last)
+				perror(opnames[op - first]);
 			else
 				perror("ioctl[unknown???]");
 		}
@@ -121,7 +150,7 @@ set80211param(struct wpa_driver_madwifi_data *drv, int op, int arg,
 
 static int
 wpa_driver_madwifi_set_wpa_ie(struct wpa_driver_madwifi_data *drv,
-			      const char *wpa_ie, size_t wpa_ie_len)
+			      const u8 *wpa_ie, size_t wpa_ie_len)
 {
 	struct iwreq iwr;
 
@@ -359,6 +388,7 @@ wpa_driver_madwifi_scan(void *priv, const u8 *ssid, size_t ssid_len)
 {
 	struct wpa_driver_madwifi_data *drv = priv;
 	struct iwreq iwr;
+	int ret = 0;
 
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
@@ -367,14 +397,24 @@ wpa_driver_madwifi_scan(void *priv, const u8 *ssid, size_t ssid_len)
 	/* FIX: scan should not break the current association, so using
 	 * set_ssid may not be the best way of doing this.. */
 	if (wpa_driver_wext_set_ssid(drv->wext, ssid, ssid_len) < 0)
-		return -1;
+		ret = -1;
 
 	if (ioctl(drv->sock, SIOCSIWSCAN, &iwr) < 0) {
 		perror("ioctl[SIOCSIWSCAN]");
-		return -1;
+		ret = -1;
 	}
-	/* NB: madwifi delivers a scan complete event so no need to poll */
-	return 0;
+
+	/*
+	 * madwifi delivers a scan complete event so no need to poll, but
+	 * register a backup timeout anyway to make sure that we recover even
+	 * if the driver does not send this event for any reason. This timeout
+	 * will only be used if the event is not delivered (event handler will
+	 * cancel the timeout).
+	 */
+	eloop_register_timeout(30, 0, wpa_driver_wext_scan_timeout, drv->wext,
+			       drv->ctx);
+
+	return ret;
 }
 
 static int wpa_driver_madwifi_get_bssid(void *priv, u8 *bssid)
