@@ -10,6 +10,12 @@
  * license.
  *
  * See README and COPYING for more details.
+ *
+ * This file implements a driver interface for the Linux Wireless Extensions.
+ * When used with WE-18 or newer, this interface can be used as-is with number
+ * of drivers. In addition to this, some of the common functions in this file
+ * can be used by other driver interface implementations that use generic WE
+ * ioctls, but require private ioctls for some of the functionality.
  */
 
 #include <stdlib.h>
@@ -36,6 +42,8 @@ struct wpa_driver_wext_data {
 	int event_sock;
 	int ioctl_sock;
 	char ifname[IFNAMSIZ + 1];
+	int ifindex;
+	int ifindex2;
 	u8 *assoc_req_ies;
 	size_t assoc_req_ies_len;
 	u8 *assoc_resp_ies;
@@ -51,20 +59,20 @@ static int wpa_driver_wext_get_range(void *priv);
 
 
 static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
-					  int index, u32 value)
+					  int idx, u32 value)
 {
 	struct iwreq iwr;
 	int ret = 0;
 
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
-	iwr.u.param.flags = index & IW_AUTH_INDEX;
+	iwr.u.param.flags = idx & IW_AUTH_INDEX;
 	iwr.u.param.value = value;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWAUTH, &iwr) < 0) {
 		perror("ioctl[SIOCSIWAUTH]");
 		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
-			index, value);
+			idx, value);
 		ret = -1;
 	}
 
@@ -72,6 +80,12 @@ static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 }
 
 
+/**
+ * wpa_driver_wext_get_bssid - Get BSSID, SIOCGIWAP
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @bssid: Buffer for BSSID
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_get_bssid(void *priv, u8 *bssid)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -91,6 +105,12 @@ int wpa_driver_wext_get_bssid(void *priv, u8 *bssid)
 }
 
 
+/**
+ * wpa_driver_wext_set_bssid - Set BSSID, SIOCSIWAP
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @bssid: BSSID
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_set_bssid(void *priv, const u8 *bssid)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -114,6 +134,12 @@ int wpa_driver_wext_set_bssid(void *priv, const u8 *bssid)
 }
 
 
+/**
+ * wpa_driver_wext_get_ssid - Get SSID, SIOCGIWESSID
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @ssid: Buffer for the SSID; must be at least 32 bytes long
+ * Returns: SSID length on success, -1 on failure
+ */
 int wpa_driver_wext_get_ssid(void *priv, u8 *ssid)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -128,23 +154,20 @@ int wpa_driver_wext_get_ssid(void *priv, u8 *ssid)
 	if (ioctl(drv->ioctl_sock, SIOCGIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCGIWESSID]");
 		ret = -1;
-	} else {
+	} else
 		ret = iwr.u.essid.length;
-		if (ret > 32)
-			ret = 32;
-		/* Some drivers include nul termination in the SSID, so let's
-		 * remove it here before further processing. WE-21 changes this
-		 * to explicitly require the length _not_ to include nul
-		 * termination. */
-		if (ret > 0 && ssid[ret - 1] == '\0' &&
-		    drv->we_version_compiled < 21)
-			ret--;
-	}
 
 	return ret;
 }
 
 
+/**
+ * wpa_driver_wext_set_ssid - Set SSID, SIOCSIWESSID
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @ssid: SSID
+ * @ssid_len: Length of SSID (0..32)
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -162,19 +185,12 @@ int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 	memset(buf, 0, sizeof(buf));
 	memcpy(buf, ssid, ssid_len);
 	iwr.u.essid.pointer = (caddr_t) buf;
-	if (drv->we_version_compiled < 21) {
-		/* For historic reasons, set SSID length to include one extra
-		 * character, C string nul termination, even though SSID is
-		 * really an octet string that should not be presented as a C
-		 * string. Some Linux drivers decrement the length by one and
-		 * can thus end up missing the last octet of the SSID if the
-		 * length is not incremented here. WE-21 changes this to
-		 * explicitly require the length _not_ to include nul
-		 * termination. */
-		if (ssid_len)
-			ssid_len++;
-	}
-	iwr.u.essid.length = ssid_len;
+	/* For historic reasons, set SSID length to include one extra
+	 * character, C string nul termination, even though SSID is really an
+	 * octet string that should not be presented as a C string. Some Linux
+	 * drivers decrement the length by one and can thus end up missing the
+	 * last octet of the SSID if the length is not incremented here. */
+	iwr.u.essid.length = ssid_len ? ssid_len + 1 : 0;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCSIWESSID]");
@@ -185,6 +201,12 @@ int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 }
 
 
+/**
+ * wpa_driver_wext_set_freq - Set frequency/channel, SIOCSIWFREQ
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @freq: Frequency in MHz
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_set_freq(void *priv, int freq)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -518,7 +540,12 @@ static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 
 	ifi = NLMSG_DATA(h);
 
-	/* TODO: use ifi->ifi_index to recognize the interface (?) */
+	if (drv->ifindex != ifi->ifi_index && drv->ifindex2 != ifi->ifi_index)
+	{
+		wpa_printf(MSG_DEBUG, "Ignore event for foreign ifindex %d",
+			   ifi->ifi_index);
+		return;
+	}
 
 	nlmsg_len = NLMSG_ALIGN(sizeof(struct ifinfomsg));
 
@@ -631,6 +658,12 @@ static void wpa_driver_wext_event_receive(int sock, void *eloop_ctx,
 }
 
 
+/**
+ * wpa_driver_wext_get_ifflags - Get interface flags (SIOCGIFFLAGS)
+ * @drv: driver_wext private data
+ * @flags: Pointer to returned flags value
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_get_ifflags(struct wpa_driver_wext_data *drv, int *flags)
 {
 	struct ifreq ifr;
@@ -646,6 +679,12 @@ int wpa_driver_wext_get_ifflags(struct wpa_driver_wext_data *drv, int *flags)
 }
 
 
+/**
+ * wpa_driver_wext_set_ifflags - Set interface flags (SIOCSIFFLAGS)
+ * @drv: driver_wext private data
+ * @flags: New value for flags
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_set_ifflags(struct wpa_driver_wext_data *drv, int flags)
 {
 	struct ifreq ifr;
@@ -661,6 +700,13 @@ int wpa_driver_wext_set_ifflags(struct wpa_driver_wext_data *drv, int flags)
 }
 
 
+/**
+ * wpa_driver_wext_init - Initialize WE driver interface
+ * @ctx: context to be used when calling wpa_supplicant functions,
+ * e.g., wpa_supplicant_event()
+ * @ifname: interface name, e.g., wlan0
+ * Returns: Pointer to private data, %NULL on failure
+ */
 void * wpa_driver_wext_init(void *ctx, const char *ifname)
 {
 	int s, flags;
@@ -719,10 +765,34 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 
 	wpa_driver_wext_get_range(drv);
 
+	drv->ifindex = if_nametoindex(drv->ifname);
+
+	if (strncmp(ifname, "wlan", 4) == 0) {
+		/*
+		 * Host AP driver may use both wlan# and wifi# interface in
+		 * wireless events. Since some of the versions included WE-18
+		 * support, let's add the alternative ifindex also from
+		 * driver_wext.c for the time being. This may be removed at
+		 * some point once it is believed that old versions of the
+		 * driver are not in use anymore.
+		 */
+		char ifname2[IFNAMSIZ + 1];
+		strncpy(ifname2, ifname, sizeof(ifname2));
+		memcpy(ifname2, "wifi", 4);
+		wpa_driver_wext_alternative_ifindex(drv, ifname2);
+	}
+
 	return drv;
 }
 
 
+/**
+ * wpa_driver_wext_deinit - Deinitialize WE driver interface
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ *
+ * Shut down driver interface and processing of driver events. Free
+ * private data buffer if one was allocated in wpa_driver_wext_init().
+ */
 void wpa_driver_wext_deinit(void *priv)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -747,6 +817,14 @@ void wpa_driver_wext_deinit(void *priv)
 }
 
 
+/**
+ * wpa_driver_wext_scan_timeout - Scan timeout to report scan completion
+ * @eloop_ctx: Unused
+ * @timeout_ctx: ctx argument given to wpa_driver_wext_init()
+ *
+ * This function can be used as registered timeout when starting a scan to
+ * generate a scan completed event if the driver does not report this.
+ */
 void wpa_driver_wext_scan_timeout(void *eloop_ctx, void *timeout_ctx)
 {
 	wpa_printf(MSG_DEBUG, "Scan timeout - try to get results");
@@ -754,6 +832,15 @@ void wpa_driver_wext_scan_timeout(void *eloop_ctx, void *timeout_ctx)
 }
 
 
+/**
+ * wpa_driver_wext_scan - Request the driver to initiate scan
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @ssid: Specific SSID to scan for (ProbeReq) or %NULL to scan for
+ *	all SSIDs (either active scan with broadcast SSID or passive
+ *	scan
+ * @ssid_len: Length of the SSID
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_scan(void *priv, const u8 *ssid, size_t ssid_len)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -829,6 +916,18 @@ static int wpa_scan_result_compar(const void *a, const void *b)
 }
 
 
+/**
+ * wpa_driver_wext_get_scan_results - Fetch the latest scan results
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @results: Pointer to buffer for scan results
+ * @max_size: Maximum number of entries (buffer size)
+ * Returns: Number of scan result entries used on success, -1 on
+ * failure
+ *
+ * If scan results include more than max_size BSSes, max_size will be
+ * returned and the remaining entries will not be included in the
+ * buffer.
+ */
 int wpa_driver_wext_get_scan_results(void *priv,
 				     struct wpa_scan_result *results,
 				     size_t max_size)
@@ -937,7 +1036,7 @@ int wpa_driver_wext_get_scan_results(void *priv,
 			break;
 		case SIOCGIWFREQ:
 			if (ap_num < max_size) {
-				int div = 1000000, i;
+				int divi = 1000000, i;
 				if (iwe->u.freq.e > 6) {
 					wpa_printf(
 						MSG_DEBUG, "Invalid freq "
@@ -948,8 +1047,8 @@ int wpa_driver_wext_get_scan_results(void *priv,
 					break;
 				}
 				for (i = 0; i < iwe->u.freq.e; i++)
-					div /= 10;
-				results[ap_num].freq = iwe->u.freq.m / div;
+					divi /= 10;
+				results[ap_num].freq = iwe->u.freq.m / divi;
 			}
 			break;
 		case IWEVQUAL:
@@ -1243,6 +1342,32 @@ static int wpa_driver_wext_set_key_ext(void *priv, wpa_alg alg,
 }
 
 
+/**
+ * wpa_driver_wext_set_key - Configure encryption key
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @priv: Private driver interface data
+ * @alg: Encryption algorithm (%WPA_ALG_NONE, %WPA_ALG_WEP,
+ *	%WPA_ALG_TKIP, %WPA_ALG_CCMP); %WPA_ALG_NONE clears the key.
+ * @addr: Address of the peer STA or ff:ff:ff:ff:ff:ff for
+ *	broadcast/default keys
+ * @key_idx: key index (0..3), usually 0 for unicast keys
+ * @set_tx: Configure this key as the default Tx key (only used when
+ *	driver does not support separate unicast/individual key
+ * @seq: Sequence number/packet number, seq_len octets, the next
+ *	packet number to be used for in replay protection; configured
+ *	for Rx keys (in most cases, this is only used with broadcast
+ *	keys and set to zero for unicast keys)
+ * @seq_len: Length of the seq, depends on the algorithm:
+ *	TKIP: 6 octets, CCMP: 6 octets
+ * @key: Key buffer; TKIP: 16-byte temporal key, 8-byte Tx Mic key,
+ *	8-byte Rx Mic Key
+ * @key_len: Length of the key buffer in octets (WEP: 5 or 13,
+ *	TKIP: 32, CCMP: 16)
+ * Returns: 0 on success, -1 on failure
+ *
+ * This function uses SIOCSIWENCODEEXT by default, but tries to use
+ * SIOCSIWENCODE if the extended ioctl fails when configuring a WEP key.
+ */
 int wpa_driver_wext_set_key(void *priv, wpa_alg alg,
 			    const u8 *addr, int key_idx,
 			    int set_tx, const u8 *seq, size_t seq_len,
@@ -1465,10 +1590,7 @@ wpa_driver_wext_associate(void *priv,
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_KEY_MGMT, value) < 0)
 		ret = -1;
-	value = params->key_mgmt_suite != KEY_MGMT_NONE ||
-		params->pairwise_suite != CIPHER_NONE ||
-		params->group_suite != CIPHER_NONE ||
-		params->wpa_ie_len;
+	value = params->key_mgmt_suite != KEY_MGMT_NONE;
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_PRIVACY_INVOKED, value) < 0)
 		ret = -1;
@@ -1519,6 +1641,12 @@ static int wpa_driver_wext_set_auth_alg(void *priv, int auth_alg)
 }
 
 
+/**
+ * wpa_driver_wext_set_mode - Set wireless mode (infra/adhoc), SIOCSIWMODE
+ * @priv: Pointer to private wext data from wpa_driver_wext_init()
+ * @mode: 0 = infra/BSS (associate with an AP), 1 = adhoc/IBSS
+ * Returns: 0 on success, -1 on failure
+ */
 int wpa_driver_wext_set_mode(void *priv, int mode)
 {
 	struct wpa_driver_wext_data *drv = priv;
@@ -1558,7 +1686,8 @@ static int wpa_driver_wext_pmksa(struct wpa_driver_wext_data *drv,
 	iwr.u.data.length = sizeof(pmksa);
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWPMKSA, &iwr) < 0) {
-		perror("ioctl[SIOCSIWPMKSA]");
+		if (errno != EOPNOTSUPP)
+			perror("ioctl[SIOCSIWPMKSA]");
 		ret = -1;
 	}
 
@@ -1599,6 +1728,25 @@ static int wpa_driver_wext_get_capa(void *priv, struct wpa_driver_capa *capa)
 }
 
 
+int wpa_driver_wext_alternative_ifindex(struct wpa_driver_wext_data *drv,
+					const char *ifname)
+{
+	if (ifname == NULL) {
+		drv->ifindex2 = -1;
+		return 0;
+	}
+
+	drv->ifindex2 = if_nametoindex(ifname);
+	if (drv->ifindex2 <= 0)
+		return -1;
+
+	wpa_printf(MSG_DEBUG, "Added alternative ifindex %d (%s) for "
+		   "wireless events", drv->ifindex2, ifname);
+
+	return 0;
+}
+
+
 const struct wpa_driver_ops wpa_driver_wext_ops = {
 	.name = "wext",
 	.desc = "Linux wireless extensions (generic)",
